@@ -249,7 +249,7 @@ parse_token_webkdc_service(struct webauth_context *ctx, const char *data,
     key_data = state->service->session_key;
     s = webauth_key_create(ctx, WA_KEY_AES, size, key_data, &key);
     if (s != WA_ERR_NONE) {
-        wai_log_error(ctx, s, WA_LOG_WARN,
+        wai_log_error(ctx, WA_LOG_WARN, s,
                       "invalid session key in webkdc-service token");
         wai_error_set(ctx, WA_PEC_SERVICE_TOKEN_INVALID, NULL);
         return WA_PEC_SERVICE_TOKEN_INVALID;
@@ -637,8 +637,15 @@ do_logins(struct webauth_context *ctx, struct wai_webkdc_login_state *state)
             s = do_login_otp(ctx, state, &token->token.login);
         else
             s = do_login_krb(ctx, state, &token->token.login);
-        if (s != WA_ERR_NONE)
+
+        /*
+         * If the login fails, store the user corresponding to the failed
+         * login token in the state struct for logging purposes.
+         */
+        if (s != WA_ERR_NONE) {
+            state->login_subject = token->token.login.username;
             return s;
+        }
     }
     return WA_ERR_NONE;
 }
@@ -1005,19 +1012,22 @@ add_user_info(struct webauth_context *ctx,
  * information (which may be NULL if there's no information service
  * configured), check whether multifactor authentication and a level of
  * assurance restriction is already satisfied or unnecessary, required, or
- * impossible.
+ * impossible.  If the request asks for a Kerberos authenticator or Kerberos
+ * proxy credentials, also check if we can satisfy those with the proxy token
+ * we have.
  *
  * Returns a WebAuth status code, and sets the configured and required factors
  * if multifactor is needed and can be met by the user's configured factors.
  */
 static int
-check_multifactor(struct webauth_context *ctx,
-                  struct wai_webkdc_login_state *state,
-                  struct webauth_user_info *info)
+check_factors_proxy(struct webauth_context *ctx,
+                    struct wai_webkdc_login_state *state,
+                    struct webauth_user_info *info)
 {
     struct webauth_factors *wanted, *swanted, *have, *shave;
     const struct webauth_factors *configured;
     struct webauth_token_webkdc_proxy *wpt;
+    bool need_kerberos;
     int s = WA_ERR_NONE;
 
     /* Figure out what factors we want and have. */
@@ -1035,9 +1045,9 @@ check_multifactor(struct webauth_context *ctx,
         wanted = webauth_factors_union(ctx, wanted, info->required);
 
     /*
-     * Second, check the level of assurance required.  If the user cannot
-     * establish a sufficient level of assurance, punt immediately; we don't
-     * care about the available factors in that case.
+     * Check the level of assurance required.  If the user cannot establish a
+     * sufficient level of assurance, punt immediately; we don't care about
+     * the available factors in that case.
      */
     if (state->request->loa > wpt->loa) {
         if (info != NULL && state->request->loa > info->max_loa) {
@@ -1048,22 +1058,51 @@ check_multifactor(struct webauth_context *ctx,
     }
 
     /*
-     * Third, see if the WAS-requested factors are already satisfied by the
-     * factors that we have.  If not, choose the error message.  If the user
-     * can't satisfy the factors at all, we'll change the error later.  Be
-     * careful not to override errors from the LoA check.
+     * If we need Kerberos, check if we have a Kerberos authenticator.  If
+     * not, set our initial error code.  We may override that later if the
+     * user can't satisfy the required factors.
+     */
+    need_kerberos = false;
+    if (strcmp(state->request->type, "id") == 0) {
+        if (strcmp(state->request->auth, "krb5") == 0)
+            need_kerberos = true;
+    } else if (strcmp(state->request->type, "proxy") == 0) {
+        if (strcmp(state->request->proxy_type, "krb5") == 0)
+            need_kerberos = true;
+    }
+    if (need_kerberos && strcmp(wpt->proxy_type, "krb5") != 0)
+        s = WA_PEC_PROXY_TOKEN_REQUIRED;
+
+    /*
+     * See if the WAS-requested factors are already satisfied by the factors
+     * that we have.  If not, choose the error message.  If the user can't
+     * satisfy the factors at all, we'll change the error later.  Be careful
+     * not to override errors from the LoA or Kerberos authenticator checks.
+     *
+     * If the user has no password session factor, we need to start them at
+     * the beginning of the login process to ensure we get all the required
+     * factors and any synthesized multifactor.  But be careful not to loop on
+     * the password screen if they do have a password session factor already.
      *
      * We assume that if the user needs factors they don't have but are
      * capable of getting, the correct next step is to force a multifactor
      * authentication.  This may not be the correct assumption always, but it
      * works for the most common cases.
      */
-    if (webauth_factors_satisfies(ctx, have, wanted)) {
-        if (!webauth_factors_satisfies(ctx, shave, swanted))
-            if (s == WA_ERR_NONE)
+    if (s == WA_ERR_NONE) {
+        if (webauth_factors_satisfies(ctx, have, wanted)) {
+            if (!webauth_factors_satisfies(ctx, shave, swanted)) {
+                if (webauth_factors_contains(ctx, shave, WA_FA_PASSWORD))
+                    s = WA_PEC_MULTIFACTOR_REQUIRED;
+                else
+                    s = WA_PEC_LOGIN_FORCED;
+            }
+        } else {
+            if (webauth_factors_contains(ctx, have, WA_FA_PASSWORD))
+                s = WA_PEC_MULTIFACTOR_REQUIRED;
+            else
                 s = WA_PEC_LOGIN_FORCED;
-    } else {
-        s = WA_PEC_MULTIFACTOR_REQUIRED;
+        }
     }
     if (s == WA_ERR_NONE)
         return WA_ERR_NONE;
@@ -1612,6 +1651,14 @@ encode_response(struct webauth_context *ctx,
     s = encode_webkdc_proxy(ctx, state->wkproxy, response, ring);
     if (s != WA_ERR_NONE)
         return s;
+
+    /*
+     * If there were no webkdc-proxy token, but we do have a subject from a
+     * failed login token, use that as the response token.  This is primarily
+     * for better logging of failures.
+     */
+    if (state->wkproxy == NULL && state->login_subject != NULL)
+        response->subject = state->login_subject;
     return WA_ERR_NONE;
 }
 
@@ -1714,8 +1761,12 @@ webauth_webkdc_login(struct webauth_context *ctx,
      * some other factor we don't have is required, reject the login with
      * either multifactor required or with multifactor unavailable, depending
      * on whether the user has multifactor configured.
+     *
+     * We also check here whether we have a Kerberos proxy token if the
+     * request asks for a Kerberos authenticator or for Kerberos proxy
+     * credentials.
      */
-    s = check_multifactor(ctx, &state, info);
+    s = check_factors_proxy(ctx, &state, info);
     if (s != WA_ERR_NONE)
         goto done;
 
@@ -1758,7 +1809,7 @@ done:
         s = result;
         result = wai_error_protocol(ctx, s);
         if (result != s) {
-            wai_log_error(ctx, s, WA_LOG_WARN, "cannot handle login request");
+            wai_log_error(ctx, WA_LOG_WARN, s, "cannot handle login request");
             wai_error_set(ctx, result, NULL);
         }
     }
